import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.mass.gasp_based.wing import (WingMassGroup,
                                                    WingMassSolve,
                                                    WingMassTotal)
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission


# this is the large single aisle 1 V3 test case
class WingMassSolveTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("wingfuel", WingMassSolve(
            aviary_options=get_option_defaults()), promotes=["*"])

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS, val=3645, units="lbm")
        self.prob.model.set_input_defaults("c_strut_braced", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_COEFFICIENT, val=102.5, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MATERIAL_FACTOR, val=1.2213063198183813, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POSITION_FACTOR, val=0.98, units="unitless")
        self.prob.model.set_input_defaults("c_gear_loc", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            "half_sweep", val=0.3947081519145335, units="rad"
        )

        newton = self.prob.model.nonlinear_solver = om.NewtonSolver()
        newton.options["atol"] = 1e-9
        newton.options["rtol"] = 1e-9
        newton.options["iprint"] = 2
        newton.options["maxiter"] = 10
        newton.options["solve_subsystems"] = True
        newton.options["max_sub_solves"] = 10
        newton.options["err_on_non_converge"] = True
        newton.options["reraise_child_analysiserror"] = False
        newton.linesearch = om.BoundsEnforceLS()
        newton.linesearch.options["bound_enforcement"] = "scalar"
        newton.linesearch.options["iprint"] = -1
        newton.options["err_on_non_converge"] = False

        self.prob.model.linear_solver = om.DirectSolver(assemble_jac=True)

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob["isolated_wing_mass"], 15830, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


# this is the large single aisle 1 V3 test case
class TotalWingMassTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "total",
            WingMassTotal(aviary_options=get_option_defaults()),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            "isolated_wing_mass", val=15830.0, units="lbm")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob[Aircraft.Wing.MASS], 15830.0, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class TotalWingMassTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "total",
            WingMassTotal(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            "isolated_wing_mass", val=15830.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=100, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDING_AREA, val=50, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLD_MASS_COEFFICIENT, val=0.2, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MASS], 17413, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class TotalWingMassTestCase3(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "total",
            WingMassTotal(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            "isolated_wing_mass", val=15830.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Strut.MASS_COEFFICIENT, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MASS], 23745, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class TotalWingMassTestCase4(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "total", WingMassTotal(aviary_options=options), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            "isolated_wing_mass", val=15830.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=100, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDING_AREA, val=50, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLD_MASS_COEFFICIENT, val=0.2, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.MASS_COEFFICIENT, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MASS], 25328, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


# this is the large single aisle 1 V3 test case
class WingMassGroupTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingMassGroup(aviary_options=get_option_defaults()),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS, val=3645, units="lbm")
        self.prob.model.set_input_defaults("c_strut_braced", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_COEFFICIENT, val=102.5, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MATERIAL_FACTOR, val=1.2213063198183813, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POSITION_FACTOR, val=0.98, units="unitless")
        self.prob.model.set_input_defaults("c_gear_loc", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            "half_sweep", val=0.3947081519145335, units="rad"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob[Aircraft.Wing.MASS], 15830, tol)
        assert_near_equal(self.prob["isolated_wing_mass"], 15830, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class WingMassGroupTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingMassGroup(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS, val=3645, units="lbm")
        self.prob.model.set_input_defaults("c_strut_braced", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_COEFFICIENT, val=102.5, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MATERIAL_FACTOR, val=1.2213063198183813, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POSITION_FACTOR, val=0.98, units="unitless")
        self.prob.model.set_input_defaults("c_gear_loc", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            "half_sweep", val=0.3947081519145335, units="rad"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=100, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDING_AREA, val=50, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLD_MASS_COEFFICIENT, val=0.2, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MASS], 17417, tol
        )  # not actual GASP value
        assert_near_equal(self.prob["isolated_wing_mass"], 15830, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class WingMassGroupTestCase3(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            WingMassGroup(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS, val=3645, units="lbm")
        self.prob.model.set_input_defaults("c_strut_braced", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_COEFFICIENT, val=102.5, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MATERIAL_FACTOR, val=1.2213063198183813, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POSITION_FACTOR, val=0.98, units="unitless")
        self.prob.model.set_input_defaults("c_gear_loc", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            "half_sweep", val=0.3947081519145335, units="rad"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Strut.MASS_COEFFICIENT, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MASS], 23750, tol
        )  # not actual GASP value
        assert_near_equal(self.prob["isolated_wing_mass"], 15830, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class WingMassGroupTestCase4(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", WingMassGroup(aviary_options=options), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS, val=3645, units="lbm")
        self.prob.model.set_input_defaults("c_strut_braced", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_COEFFICIENT, val=102.5, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MATERIAL_FACTOR, val=1.2213063198183813, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POSITION_FACTOR, val=0.98, units="unitless")
        self.prob.model.set_input_defaults("c_gear_loc", val=1.0, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            "half_sweep", val=0.3947081519145335, units="rad"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=100, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDING_AREA, val=50, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLD_MASS_COEFFICIENT, val=0.2, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.MASS_COEFFICIENT, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MASS], 25333, tol
        )  # not actual GASP value
        assert_near_equal(self.prob["isolated_wing_mass"], 15830, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
import subprocess

from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs

from aviary.interface.methods_for_level1 import run_aviary
from aviary.subsystems.test.test_dummy_subsystem import ArrayGuessSubsystemBuilder
from aviary.mission.flops_based.phases.energy_phase import EnergyPhase
from aviary.variable_info.variables import Dynamic


@use_tempdirs
class AircraftMissionTestSuite(unittest.TestCase):

    def setUp(self):

        # Load the phase_info and other common setup tasks
        self.phase_info = {
            "pre_mission": {"include_takeoff": False, "optimize_mass": True},
            "climb": {
                "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
                "user_options": {
                    "optimize_mach": False,
                    "optimize_altitude": False,
                    "polynomial_control_order": 1,
                    "num_segments": 5,
                    "order": 3,
                    "solve_for_distance": False,
                    "initial_mach": (0.2, "unitless"),
                    "final_mach": (0.72, "unitless"),
                    "mach_bounds": ((0.18, 0.74), "unitless"),
                    "initial_altitude": (0.0, "ft"),
                    "final_altitude": (32000.0, "ft"),
                    "altitude_bounds": ((0.0, 34000.0), "ft"),
                    "throttle_enforcement": "path_constraint",
                    "fix_initial": True,
                    "constrain_final": False,
                    "fix_duration": False,
                    "initial_bounds": ((0.0, 0.0), "min"),
                    "duration_bounds": ((64.0, 192.0), "min"),
                },
                "initial_guesses": {"times": ([0, 128], "min")},
            },
            "cruise": {
                "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
                "user_options": {
                    "optimize_mach": False,
                    "optimize_altitude": False,
                    "polynomial_control_order": 1,
                    "num_segments": 5,
                    "order": 3,
                    "solve_for_distance": False,
                    "initial_mach": (0.72, "unitless"),
                    "final_mach": (0.72, "unitless"),
                    "mach_bounds": ((0.7, 0.74), "unitless"),
                    "initial_altitude": (32000.0, "ft"),
                    "final_altitude": (34000.0, "ft"),
                    "altitude_bounds": ((23000.0, 38000.0), "ft"),
                    "throttle_enforcement": "boundary_constraint",
                    "fix_initial": False,
                    "constrain_final": False,
                    "fix_duration": False,
                    "initial_bounds": ((64.0, 192.0), "min"),
                    "duration_bounds": ((56.5, 169.5), "min"),
                },
                "initial_guesses": {"times": ([128, 113], "min")},
            },
            "descent": {
                "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
                "user_options": {
                    "optimize_mach": False,
                    "optimize_altitude": False,
                    "polynomial_control_order": 1,
                    "num_segments": 5,
                    "order": 3,
                    "solve_for_distance": False,
                    "initial_mach": (0.72, "unitless"),
                    "final_mach": (0.36, "unitless"),
                    "mach_bounds": ((0.34, 0.74), "unitless"),
                    "initial_altitude": (34000.0, "ft"),
                    "final_altitude": (500.0, "ft"),
                    "altitude_bounds": ((0.0, 38000.0), "ft"),
                    "throttle_enforcement": "path_constraint",
                    "fix_initial": False,
                    "constrain_final": True,
                    "fix_duration": False,
                    "initial_bounds": ((120.5, 361.5), "min"),
                    "duration_bounds": ((29.0, 87.0), "min"),
                },
                "initial_guesses": {"times": ([241, 58], "min")},
            },
            "post_mission": {
                "include_landing": False,
                "constrain_range": True,
                "target_range": (1906, "nmi"),
            },
        }

        self.aircraft_definition_file = 'models/test_aircraft/aircraft_for_bench_FwFm.csv'
        self.make_plots = False
        self.max_iter = 100

    def add_external_subsystem(self, phase_info, subsystem_builder):
        """
        Add an external subsystem to all phases in the mission.
        """
        for phase in phase_info:
            if 'user_options' in phase_info[phase]:
                if 'external_subsystems' not in phase_info[phase]:
                    phase_info[phase]['external_subsystems'] = []
                phase_info[phase]['external_subsystems'].append(subsystem_builder)

    def run_mission(self, phase_info, optimizer):
        return run_aviary(
            self.aircraft_definition_file, phase_info,
            make_plots=self.make_plots, max_iter=self.max_iter, optimizer=optimizer,
            optimization_history_filename="driver_test.db")

    def test_mission_basic_and_dashboard(self):
        # We need to remove the TESTFLO_RUNNING environment variable for this test to run.
        # The reports code checks to see if TESTFLO_RUNNING is set and will not do anything if set
        # But we need to remember whether it was set so we can restore it
        testflo_running = os.environ.pop('TESTFLO_RUNNING', None)

        prob = self.run_mission(self.phase_info, "SLSQP")

        # restore what was there before running the test
        if testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = testflo_running

        self.assertIsNotNone(prob)
        self.assertFalse(prob.failed)

        cmd = f'aviary dashboard --problem_recorder dymos_solution.db --driver_recorder driver_test.db {prob.driver._problem()._name}'
        # this only tests that a given command line tool returns a 0 return code. It doesn't
        # check the expected output at all.  The underlying functions that implement the
        # commands should be tested seperately.
        try:
            subprocess.check_output(cmd.split())
        except subprocess.CalledProcessError as err:
            self.fail("Command '{}' failed.  Return code: {}".format(cmd, err.returncode))

    @require_pyoptsparse(optimizer="IPOPT")
    def test_mission_basic_pyopt(self):
        prob = self.run_mission(self.phase_info, "IPOPT")
        self.assertIsNotNone(prob)
        self.assertFalse(prob.failed)

    @require_pyoptsparse(optimizer="IPOPT")
    def test_mission_optimize_mach_only(self):
        # Test with optimize_mach flag set to True
        modified_phase_info = self.phase_info.copy()
        for phase in ["climb", "cruise", "descent"]:
            modified_phase_info[phase]["user_options"]["optimize_mach"] = True
        prob = self.run_mission(modified_phase_info, "IPOPT")
        self.assertFalse(prob.failed)

    @require_pyoptsparse(optimizer="IPOPT")
    def test_mission_optimize_altitude_and_mach(self):
        # Test with optimize_altitude flag set to True
        modified_phase_info = self.phase_info.copy()
        for phase in ["climb", "cruise", "descent"]:
            modified_phase_info[phase]["user_options"]["optimize_altitude"] = True
            modified_phase_info[phase]["user_options"]["optimize_mach"] = True
        modified_phase_info['climb']['user_options']['constraints'] = {
            Dynamic.Mission.THROTTLE: {
                'lower': 0.2,
                'upper': 0.9,
                'type': 'path',
            },
        }
        prob = self.run_mission(modified_phase_info, "IPOPT")
        self.assertFalse(prob.failed)

        constraints = prob.driver._cons
        for name, meta in constraints.items():
            if 'traj.phases.climb->path_constraint->throttle' in name:
                self.assertEqual(meta['upper'], 0.9)
                self.assertEqual(meta['lower'], 0.2)

    @require_pyoptsparse(optimizer="IPOPT")
    def test_mission_optimize_altitude_only(self):
        # Test with optimize_altitude flag set to True
        modified_phase_info = self.phase_info.copy()
        for phase in ["climb", "cruise", "descent"]:
            modified_phase_info[phase]["user_options"]["optimize_altitude"] = True
        prob = self.run_mission(modified_phase_info, "IPOPT")
        self.assertFalse(prob.failed)

    @require_pyoptsparse(optimizer="IPOPT")
    def test_mission_solve_for_distance(self):
        modified_phase_info = self.phase_info.copy()
        for phase in ["climb", "cruise", "descent"]:
            modified_phase_info[phase]["user_options"]["solve_for_distance"] = True
        prob = self.run_mission(modified_phase_info, "IPOPT")
        self.assertFalse(prob.failed)

    def test_mission_solve_for_distance(self):
        modified_phase_info = self.phase_info.copy()
        for phase in ["climb", "cruise", "descent"]:
            modified_phase_info[phase]["user_options"]["solve_for_distance"] = True
        prob = self.run_mission(modified_phase_info, "SLSQP")
        self.assertFalse(prob.failed)

    @require_pyoptsparse(optimizer="IPOPT")
    def test_mission_with_external_subsystem(self):
        # Test mission with an external subsystem added to each phase
        modified_phase_info = self.phase_info.copy()
        dummy_subsystem_builder = ArrayGuessSubsystemBuilder()
        self.add_external_subsystem(modified_phase_info, dummy_subsystem_builder)

        prob = self.run_mission(modified_phase_info, "IPOPT")
        self.assertFalse(prob.failed)

    def test_custom_phase_builder(self):
        local_phase_info = self.phase_info.copy()
        local_phase_info['climb']['phase_builder'] = EnergyPhase

        run_aviary(self.aircraft_definition_file,
                   local_phase_info, max_iter=1, optimizer='SLSQP')

    def test_custom_phase_builder_error(self):
        local_phase_info = self.phase_info.copy()
        local_phase_info['climb']['phase_builder'] = "fake phase object"

        with self.assertRaises(TypeError):
            run_aviary(self.aircraft_definition_file,
                       local_phase_info, max_iter=1, optimizer='SLSQP')


if __name__ == '__main__':
    unittest.main()



import unittest
import os

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.ode.constraints.flight_constraints import \
    FlightConstraints
from aviary.variable_info.variables import Aircraft, Dynamic


class FlightConstraintTestCase(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", FlightConstraints(num_nodes=2), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Dynamic.Mission.MASS, np.array([174878.0, 174878.0]), units="lbm"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.AREA, 1370.3, units="ft**2")
        self.prob.model.set_input_defaults(
            "rho", 0.0023081 * np.ones(2), units="slug/ft**3"
        )
        self.prob.model.set_input_defaults(
            "CL_max", 1.2596 * np.ones(2), units="unitless")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.FLIGHT_PATH_ANGLE, 7.76 * np.ones(2), units="deg")
        self.prob.model.set_input_defaults(Aircraft.Wing.INCIDENCE, 0.0, units="deg")
        self.prob.model.set_input_defaults("alpha", 5.19 * np.ones(2), units="deg")
        self.prob.model.set_input_defaults("TAS", 252 * np.ones(2), units="kn")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob["theta"], np.array([0.2260201, 0.2260201]), tol
        )  # from GASP
        assert_near_equal(
            self.prob["TAS_violation"], np.array([-99.39848181, -99.39848181]), tol
        )  # note: output value isn't in GASP

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=3e-11, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



'''
Define utilities for using named values with associated units.

Utilities
---------
Units : type alias
    define a type hint for associated units

ValueAndUnits : type alias
    define a type hint for a single value paired with its associated units

OptionalValueAndUnits : type alias
    define a type hint for an optional single value paired with its associated units

class NamedValues
    define a collection of named values with associated units
'''
import copy
from collections.abc import Collection
from typing import Any, Tuple, Union

from openmdao.core.constants import _UNDEFINED
from openmdao.utils.units import convert_units as _convert_units

Units = str
ValueAndUnits = Tuple[Any, Units]
OptionalValueAndUnits = Union[ValueAndUnits, Any]


class NamedValues(Collection):
    '''
    Define a collection of named values with associated units.
    '''

    def __init__(self, other=None, **kwargs):
        '''
        Initialize this collection.

        Notes
        -----
        When intializing from another collection, the following types of
        collections are supported:

            * `NamedValues`
            * `Dict[str, ValueAndUnits]`
            * `Iterable[Tuple[str, ValueAndUnits]]`

        When initializing from keyword arguments, the mapped item must be of a
        type of `ValueAndUnits`.
        '''
        self._mapping = {}

        self.update(other, **kwargs)

    def get_item(self, key, default=(None, None)) -> OptionalValueAndUnits:
        '''
        Return the named value and its associated units.

        Note, this method never raises `KeyError` or `TypeError`.

        Parameters
        ----------
        key : str
            the name of the item

        default : OptionalValueAndUnits (None, None)
            if the item does not exist, return this object

        Returns
        -------
        OptionalValueAndUnits

        See Also
        --------
        get_val
        set_val
        '''
        item = self._mapping.get(key, _UNDEFINED)

        if item is _UNDEFINED:
            return default

        return item

    def copy(self):
        '''
        Return a copy of the instance of this class.

        Parameters
        ---------
        None

        Returns
        -------
        NamedValues()
        '''
        return copy.copy(self)

    def deepcopy(self):
        '''
        Return a deep copy of the instance of this class.

        Parameters
        ---------
        None

        Returns
        -------
        NamedValues()
        '''
        return copy.deepcopy(self)

    def get_val(self, key, units='unitless') -> Any:
        '''
        Return the named value in the specified units.

        Note, requesting a named value that does not exist will raise `KeyError`.

        Note, specifying units of `None` or units of any type other than `str` will raise
        `TypeError`.

        Parameters
        ----------
        key : str
            the name of the item

        units : str ('unitless')
            the units of the returned value

        Returns
        -------
        val

        Raises
        ------
        KeyError
            if the named value does not exist

        TypeError
            if units of `None` were specified or units of any type other than `str`
        '''
        self._check_units('get_val', key, units)

        item = self._mapping.get(key, _UNDEFINED)

        if item is _UNDEFINED:
            raise KeyError(f'KeyError: key not found: {key}')

        val, old_units = item

        if isinstance(val, tuple):
            val = tuple(_convert_units(v, old_units, units) for v in val)
        elif old_units != units:
            val = _convert_units(val, old_units, units)

        return val

    def set_val(self, key, val, units='unitless'):
        '''
        Update the named value and its associated units.

        Note, specifying units of `None` or units of any type other than `str` will raise
        `Typerror`.

        Parameters
        ----------
        key : str
            the name of the item

        val : Any
            the new value of the item

        units : str ('unitless')
            the units associated with the new value, if any

        Raises
        ------
        TypeError
            if units of `None` were specified or units of any type other than `str`
        '''
        self._check_units('set_val', key, units)

        self._mapping[key] = (val, units)

    def __repr__(self):
        '''
        Return a string containing a printable representation of the collection.
        '''
        return repr(self._mapping)

    def clear(self):
        '''
        Remove all items from the collection.
        '''
        self._mapping.clear()

    def update(self, other=None, **kwargs):
        '''
        Assign named values and their associated units found in another
        collection to this collection, overwriting existing items.

        If keyword arguments are specified, the collection is then assigned those
        named values and their associated units, overwriting existing items.

        Parameters
        ----------
        other (None)
            a collection of named values and their associated units

        **kwargs (optional)
            individual named values and their associated units

        Notes
        -----
        The following types of collections are supported
            * `NamedValues`
            * `Dict[str, ValueAndUnits]`
            * `Iterable[Tuple[str, ValueAndUnits]]`

        When assigning from keyword arguments, the mapped item must be of a
        type of `ValueAndUnits`.
        '''
        if not (other or kwargs):
            return

        set_val = self.set_val

        if isinstance(other, type(self)):
            # NamedValues
            other = other._mapping

        if other is not None:
            # check for dictionary
            keys = getattr(other, 'keys', None)

            if keys is None:
                # iterable, but not dictionary
                for key, (val, units) in other:
                    set_val(key, val, units)

            else:
                # dictionary
                for key in keys():
                    val, units = other[key]
                    set_val(key, val, units)

        for key, (val, units) in kwargs.items():
            set_val(key, val, units)

    def delete(self, key):
        '''
        Remove the named value and its associated units.

        Raises
        ------
        KeyError
            if the named value does not exist
        '''
        try:
            del self._mapping[key]

        except KeyError:
            raise KeyError(f'KeyError: key not found: {key}')

    def __eq__(self, other):
        '''
        Return whether or not this collection is equivalent to another.
        '''
        collection = self._mapping

        if isinstance(other, type(self)):
            return collection == other._mapping

        return collection == other

    def __contains__(self, key):
        '''
        Return whether or not the named value exists.
        '''
        return key in self._mapping

    def __iter__(self):
        '''
        Return an iterator over the `(key, (val, units))` data stored in this collection.
        '''
        items = self._mapping.items()

        yield from items

    def __len__(self):
        '''
        Return the number of items in this collection.
        '''
        return len(self._mapping)

    def _check_units(self, funcname, key, units):
        '''
        If units of `None` were specified or units of any type other than `str`, raise
        `TypeError`. Otherwise, do nothing.

        Parameters
        ----------
        funcname : str
            the name of a method/function

        key : str
            the name of the item

        units : Any
            the units to check
        '''
        if ((units is None) or not isinstance(units, str)):
            raise TypeError(
                f'{self.__class__.__name__}: {funcname}({key}):'
                f' unsupported units: {units}'
            )

    __slots__ = ('_mapping',)


def get_keys(named_values: NamedValues):
    '''
    Return a new view of the collection's names.
    '''
    return named_values._mapping.keys()


def get_items(named_values: NamedValues):
    '''
    Return a new view of the collection's `(key, (val, units))`.
    '''
    return named_values._mapping.items()


def get_values(named_values: NamedValues):
    '''
    Return a new view of the collection's `(val, units)`.
    '''
    return named_values._mapping.values()



"""
Curated list of aviary inputs that are promoted as parameters in the mission.
"""
from aviary.variable_info.variables import Aircraft, Mission


core_mission_inputs = [
    Aircraft.Design.BASE_AREA,
    Aircraft.Design.DRAG_POLAR,
    Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR,
    Aircraft.Design.LIFT_POLAR,
    Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR,
    Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR,
    Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR,
    Aircraft.Engine.SCALE_FACTOR,
    Aircraft.Fuselage.CHARACTERISTIC_LENGTH,
    Aircraft.Fuselage.CROSS_SECTION,
    Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
    Aircraft.Fuselage.FINENESS,
    Aircraft.Fuselage.LAMINAR_FLOW_LOWER,
    Aircraft.Fuselage.LAMINAR_FLOW_UPPER,
    Aircraft.Fuselage.LENGTH_TO_DIAMETER,
    Aircraft.Fuselage.WETTED_AREA,
    Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
    Aircraft.HorizontalTail.FINENESS,
    Aircraft.HorizontalTail.LAMINAR_FLOW_LOWER,
    Aircraft.HorizontalTail.LAMINAR_FLOW_UPPER,
    Aircraft.HorizontalTail.WETTED_AREA,
    Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
    Aircraft.Nacelle.FINENESS,
    Aircraft.Nacelle.LAMINAR_FLOW_LOWER,
    Aircraft.Nacelle.LAMINAR_FLOW_UPPER,
    Aircraft.Nacelle.WETTED_AREA,
    Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
    Aircraft.VerticalTail.FINENESS,
    Aircraft.VerticalTail.LAMINAR_FLOW_LOWER,
    Aircraft.VerticalTail.LAMINAR_FLOW_UPPER,
    Aircraft.VerticalTail.WETTED_AREA,
    Aircraft.Wing.AREA,
    Aircraft.Wing.ASPECT_RATIO,
    Aircraft.Wing.CHARACTERISTIC_LENGTH,
    Aircraft.Wing.FINENESS,
    Aircraft.Wing.LAMINAR_FLOW_LOWER,
    Aircraft.Wing.LAMINAR_FLOW_UPPER,
    Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
    Aircraft.Wing.SPAN_EFFICIENCY_FACTOR,
    Aircraft.Wing.SWEEP,
    Aircraft.Wing.TAPER_RATIO,
    Aircraft.Wing.THICKNESS_TO_CHORD,
    Aircraft.Wing.WETTED_AREA,
    Mission.Design.GROSS_MASS,
    Mission.Design.LIFT_COEFFICIENT,
    Mission.Design.MACH,
    Mission.Takeoff.FINAL_MASS,
]



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import (
    distributed_engine_count_factor, distributed_nacelle_diam_factor)
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class TransportStarterMass(om.ExplicitComponent):
    '''
    Calculates total sum of all engine starter masses for the entire propulsion
    system (all engines).  The methodology is based on the
    FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        # TODO nacelle outputs are vectorized
        add_aviary_input(self, Aircraft.Nacelle.AVG_DIAMETER, val=0.0)

        add_aviary_output(self, Aircraft.Propulsion.TOTAL_STARTER_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        total_engines = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_engines = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)

        d_nacelle = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)
        num_engines_factor = distributed_engine_count_factor(total_engines)
        f_nacelle = distributed_nacelle_diam_factor(d_nacelle, num_engines)

        outputs[Aircraft.Propulsion.TOTAL_STARTER_MASS] = (
            11.0 * num_engines_factor * max_mach**0.32 * f_nacelle**1.6) / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        total_engines = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_engines = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)

        d_nacelle = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        eng_count_factor = distributed_engine_count_factor(total_engines)
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        d_avg = sum(d_nacelle * num_engines) / total_engines

        diam_deriv_fact = 1
        if total_engines > 4:
            diam_deriv_fact = (0.5 * total_engines ** 0.5)**1.6

        diam_exp = diam_deriv_fact * d_avg**1.6
        max_mach_exp = max_mach**0.32

        J[Aircraft.Propulsion.TOTAL_STARTER_MASS, Aircraft.Nacelle.AVG_DIAMETER] = \
            17.6 * eng_count_factor * max_mach_exp * diam_deriv_fact * d_avg**0.6 / GRAV_ENGLISH_LBM



import openmdao.api as om
import numpy as np
from aviary.constants import GRAV_METRIC_FLOPS as gravity
from aviary.variable_info.variables import Dynamic


class RequiredThrust(om.ExplicitComponent):
    """
    Computes the required thrust using the equation:
    thrust_required = drag + (altitude_rate*gravity/velocity + velocity_rate) * mass
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(Dynamic.Mission.DRAG, val=np.zeros(nn),
                       units='N', desc='drag force')
        self.add_input(Dynamic.Mission.ALTITUDE_RATE, val=np.zeros(nn),
                       units='m/s', desc='rate of change of altitude')
        self.add_input(Dynamic.Mission.VELOCITY, val=np.zeros(nn),
                       units='m/s', desc=Dynamic.Mission.VELOCITY)
        self.add_input(Dynamic.Mission.VELOCITY_RATE, val=np.zeros(
            nn), units='m/s**2', desc='rate of change of velocity')
        self.add_input(Dynamic.Mission.MASS, val=np.zeros(
            nn), units='kg', desc='mass of the aircraft')
        self.add_output('thrust_required', val=np.zeros(
            nn), units='N', desc='required thrust')

        ar = np.arange(nn)
        self.declare_partials('thrust_required', Dynamic.Mission.DRAG, rows=ar, cols=ar)
        self.declare_partials(
            'thrust_required', Dynamic.Mission.ALTITUDE_RATE, rows=ar, cols=ar)
        self.declare_partials(
            'thrust_required', Dynamic.Mission.VELOCITY, rows=ar, cols=ar)
        self.declare_partials(
            'thrust_required', Dynamic.Mission.VELOCITY_RATE, rows=ar, cols=ar)
        self.declare_partials('thrust_required', Dynamic.Mission.MASS, rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        drag = inputs[Dynamic.Mission.DRAG]
        altitude_rate = inputs[Dynamic.Mission.ALTITUDE_RATE]
        velocity = inputs[Dynamic.Mission.VELOCITY]
        velocity_rate = inputs[Dynamic.Mission.VELOCITY_RATE]
        mass = inputs[Dynamic.Mission.MASS]

        thrust_required = drag + (altitude_rate*gravity/velocity + velocity_rate) * mass

        outputs['thrust_required'] = thrust_required

    def compute_partials(self, inputs, partials):
        altitude_rate = inputs[Dynamic.Mission.ALTITUDE_RATE]
        velocity = inputs[Dynamic.Mission.VELOCITY]
        velocity_rate = inputs[Dynamic.Mission.VELOCITY_RATE]
        mass = inputs[Dynamic.Mission.MASS]

        partials['thrust_required', Dynamic.Mission.DRAG] = 1.0
        partials['thrust_required', Dynamic.Mission.ALTITUDE_RATE] = gravity/velocity * mass
        partials['thrust_required', Dynamic.Mission.VELOCITY] = - \
            altitude_rate*gravity/velocity**2 * mass
        partials['thrust_required', Dynamic.Mission.VELOCITY_RATE] = mass
        partials['thrust_required', Dynamic.Mission.MASS] = altitude_rate * \
            gravity/velocity + velocity_rate



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.air_conditioning import (
    AltAirCondMass, TransportAirCondMass)
from aviary.subsystems.propulsion.propulsion_premission import PropulsionPreMission
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (Version,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission


class TransportAirCondMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "air_cond",
            TransportAirCondMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.AirConditioning.MASS_SCALER,
                        Aircraft.Avionics.MASS,
                        Aircraft.Fuselage.MAX_HEIGHT,
                        Aircraft.Fuselage.PLANFORM_AREA],
            output_keys=Aircraft.AirConditioning.MASS,
            aviary_option_keys=[Aircraft.CrewPayload.NUM_PASSENGERS],
            version=Version.TRANSPORT,
            tol=3.0e-4,
            atol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class AltAirCondMassTest(unittest.TestCase):
    '''
    Tests alternate air conditioning mass calculation.
    '''

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'air_cond',
            AltAirCondMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=Aircraft.AirConditioning.MASS_SCALER,
            output_keys=Aircraft.AirConditioning.MASS,
            aviary_option_keys=Aircraft.CrewPayload.NUM_PASSENGERS,
            version=Version.ALTERNATE)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.variable_info.variables import Dynamic


class RangeComp(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_nodes", types=int)

    def setup(self):

        nn = self.options["num_nodes"]

        self.add_input("cruise_time_initial", val=0.0, units="s",
                       desc="time at which cruise begins")
        self.add_input("cruise_distance_initial", val=0.0, units="NM",
                       desc="range reference at which cruise begins")

        self.add_input("TAS_cruise", val=0.8 * np.ones(nn),
                       units="ft/s", desc="true airspeed")
        self.add_input("mass", val=150000 * np.ones(nn), units="lbm",
                       desc="mass at each node, monotonically nonincreasing")

        self.add_input(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
                       0.74 * np.ones(nn), units="lbm/h")

        self.add_output("cruise_time", shape=(nn,), units="s", desc="time in cruise",
                        tags=['dymos.state_source:cruise_time'])
        self.add_output("cruise_range", shape=(nn,), units="NM", desc="cruise range",
                        tags=["dymos.state_source:distance"])

    def setup_partials(self):
        nn = self.options["num_nodes"]

        # The nonzero partials of the change in range between each two nodes (dr) wrt fuel flow and mass are
        # along two diagonals. The lower diagonal contains the partials wrt the initial values of mass or fuel
        # flow across each two-node section. The upper diagonal contains the partials wrt the final values of mass
        # or fuel flow across each two-node section. For instance, for five nodes we have:
        #
        #           0  0  0  0  0
        #           i  f  0  0  0
        # ddR/dW =  0  i  f  0  0
        #           0  0  i  f  0
        #           0  0  0  i  f
        #
        # The change of range and time at the first node is zero, and it has no dependence on the mass
        # history or fuel flow setting.
        # Since dR is the difference in range between two nodes, we need to accumulate it across all provided masses
        # using np.cumsum.
        # The partial derivative of np.cumsum is just a lower triangular matrix of ones.
        # Thus, the derivative of the accumulated range (and time) will be the matrix product:
        #           1  0  0  0  0     0  0  0  0  0
        #           1  1  0  0  0     i  f  0  0  0
        # ddR/dW =  1  1  1  0  0  @  0  i  f  0  0
        #           1  1  1  1  0     0  0  i  f  0
        #           1  1  1  1  1     0  0  0  i  f
        rs, cs = np.tril_indices(nn)
        self._tril_rs, self._tril_cs = rs, cs

        self.declare_partials(
            "cruise_range", [Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, "mass", "TAS_cruise"], rows=rs, cols=cs)
        self.declare_partials(
            "cruise_time", [Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, "mass", "TAS_cruise"], rows=rs, cols=cs)

        self.declare_partials("cruise_range", "cruise_distance_initial", val=1.0)
        self.declare_partials("cruise_time", "cruise_time_initial", val=1.0)

        # Allocated memory so we don't have to repeatedly do it in compute_partials
        # Note: since these are only used in compute_partials we don't have to worry about them supporting
        # complex values under complex step.
        self._d_cumsum_dx = np.tri(nn)
        # Note: We could make this sparse, probably doesn't matter.
        self._scratch_nn_x_nn = np.zeros((nn, nn))

    def compute(self, inputs, outputs):
        v_x = inputs["TAS_cruise"]
        m = inputs["mass"]
        FF = -inputs[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL]
        r0 = inputs["cruise_distance_initial"]
        t0 = inputs["cruise_time_initial"]

        FF_1 = FF[:-1]  # Initial fuel flow across each two-node pair
        FF_2 = FF[1:]  # Final fuel flow across each two-node pair

        # Initial weight across each two-node pair
        W1 = m[:-1] * GRAV_ENGLISH_LBM
        # Final weight across each two-node pair
        W2 = m[1:] * GRAV_ENGLISH_LBM

        vx_1 = v_x[:-1]  # Initial airspeed across each two-node pair
        vx_2 = v_x[1:]  # Final airspeed across each two-node pair
        vx_m = (vx_1 + vx_2) / 2  # Average airspeed across each two-node pair.

        breg_1 = vx_1 * W1 * 3600 / (FF_1 * 6076.1)
        breg_2 = vx_2 * W2 * 3600 / (FF_2 * 6076.1)
        bregA = (breg_1 + breg_2) / 2

        drange_cruise = bregA * np.log(1. / (1. - (W1 - W2) / W1))

        outputs["cruise_range"][0] = r0
        outputs["cruise_range"][1:] = r0 + np.cumsum(drange_cruise)
        outputs["cruise_time"][0] = t0
        outputs["cruise_time"][1:] = t0 + np.cumsum(drange_cruise) / vx_m * 6076.1

    def compute_partials(self, inputs, J):
        v_x = inputs["TAS_cruise"]
        vx_1 = v_x[:-1]  # Initial airspeed across each two-node pair
        vx_2 = v_x[1:]  # Final airspeed across each two-node pair
        vx_m = (vx_1 + vx_2) / 2  # Average airspeed across each two-node pair.

        m = inputs["mass"]
        # Initial mass across each two-node pair
        W1 = m[:-1] * GRAV_ENGLISH_LBM
        W2 = m[1:] * GRAV_ENGLISH_LBM  # Final mass across each two-node pair

        FF = -inputs[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL]
        FF_1 = FF[:-1]  # Initial fuel flow across each two-node pair
        FF_2 = FF[1:]  # Final fuel flow across each two_node pair

        breg_1 = vx_1 * W1 * 3600 / (FF_1 * 6076.1)
        breg_2 = vx_2 * W2 * 3600 / (FF_2 * 6076.1)
        bregA = (breg_1 + breg_2) / 2
        star = np.log(1 / (1 - (W1 - W2) / W1))

        dBreg1_dVx1 = W1 * 3600 / (FF_1 * 6076.1)
        dBreg1_dW1 = vx_1 * 3600 / (FF_1 * 6076.1)
        dBreg1_dFF1 = vx_1 * W1 * 3600 / (FF_1**2 * 6076.1)

        dBreg2_dVx2 = W2 * 3600 / (FF_2 * 6076.1)
        dBreg2_dW2 = vx_2 * 3600 / (FF_2 * 6076.1)
        dBreg2_dFF2 = vx_2 * W2 * 3600 / (FF_2**2 * 6076.1)

        dStar_dW1 = 1.0 / W1
        dStar_dW2 = -1.0 / W2

        dBregA_dVx1 = dBreg1_dVx1 / 2
        dBregA_dVx2 = dBreg2_dVx2 / 2
        dBregA_dW1 = dBreg1_dW1 / 2
        dBregA_dW2 = dBreg2_dW2 / 2
        dBregA_dFF1 = dBreg1_dFF1 / 2
        dBregA_dFF2 = dBreg2_dFF2 / 2

        dRange_dVx1 = dBregA_dVx1 * star
        dRange_dVx2 = dBregA_dVx2 * star
        dRange_dW1 = dBregA_dW1 * star + bregA * dStar_dW1
        dRange_dW2 = dBregA_dW2 * star + bregA * dStar_dW2
        dRange_dFF1 = dBregA_dFF1 * star
        dRange_dFF2 = dBregA_dFF2 * star

        # Partials of cruise_range

        # WRT Fuel Flow
        np.fill_diagonal(self._scratch_nn_x_nn[1:, :-1], dRange_dFF1)
        np.fill_diagonal(self._scratch_nn_x_nn[1:, 1:], dRange_dFF2)

        J["cruise_range", Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL][...] = \
            (self._d_cumsum_dx @ self._scratch_nn_x_nn)[self._tril_rs, self._tril_cs]

        # WRT Mass: dRange_dm = dRange_dW * dW_dm
        np.fill_diagonal(self._scratch_nn_x_nn[1:, :-1],
                         dRange_dW1 * GRAV_ENGLISH_LBM)
        np.fill_diagonal(self._scratch_nn_x_nn[1:, 1:],
                         dRange_dW2 * GRAV_ENGLISH_LBM)

        J["cruise_range", "mass"][...] = \
            (self._d_cumsum_dx @ self._scratch_nn_x_nn)[self._tril_rs, self._tril_cs]

        # WRT TAS_cruise
        np.fill_diagonal(self._scratch_nn_x_nn[1:, :-1], dRange_dVx1)
        np.fill_diagonal(self._scratch_nn_x_nn[1:, 1:], dRange_dVx2)

        J["cruise_range", "TAS_cruise"][...] = \
            (self._d_cumsum_dx @ self._scratch_nn_x_nn)[self._tril_rs, self._tril_cs]

        # Partials of cruise_time

        # Here we need to multiply rows [1:] of the jacobian by (6076.1 / vx_m)
        # But the jacobian is in a flat format in row-major order. The rows associated
        # with the nonzero elements are stored in self._tril_rs.

        J["cruise_time", Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL][1:] = \
            J["cruise_range", Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL][1:] / \
            vx_m[self._tril_rs[1:] - 1] * 6076.1
        J["cruise_time", "mass"][1:] = \
            J["cruise_range", "mass"][1:] / vx_m[self._tril_rs[1:] - 1] * 6076.1

        drange_cruise = bregA * star

        f = np.cumsum(drange_cruise)[:, np.newaxis]
        df_du = self._d_cumsum_dx @ self._scratch_nn_x_nn

        np.fill_diagonal(self._scratch_nn_x_nn[1:, :-1], 0.5)
        np.fill_diagonal(self._scratch_nn_x_nn[1:, 1:], 0.5)

        g = vx_m[:, np.newaxis]
        dg_du = self._scratch_nn_x_nn

        dt_dvx = 6076.1 * ((df_du[1:, ...] * g) - (dg_du[1:, ...] * f)) / g**2

        J["cruise_time", "TAS_cruise"][1:] = \
            dt_dvx[self._tril_rs[1:] - 1, self._tril_cs[1:]]



import openmdao.api as om

from aviary.subsystems.aerodynamics.gasp_based.flaps_model.basic_calculations import \
    BasicFlapsCalculations
from aviary.subsystems.aerodynamics.gasp_based.flaps_model.Cl_max import \
    CLmaxCalculation
from aviary.subsystems.aerodynamics.gasp_based.flaps_model.L_and_D_increments import \
    LiftAndDragIncrements
from aviary.subsystems.aerodynamics.gasp_based.flaps_model.meta_model import \
    MetaModelGroup
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.enums import FlapType
from aviary.variable_info.variables import Aircraft, Dynamic


class FlapsGroup(om.Group):
    """
    Group connecting four components of the flaps model.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

        # optimum trailing edge flap deflection angle defaults (ADELTO table in GASP)
        self.optimum_flap_defls = {
            FlapType.PLAIN: 60.0,
            FlapType.SPLIT: 60.0,
            FlapType.SINGLE_SLOTTED: 40.0,
            FlapType.DOUBLE_SLOTTED: 55.0,
            FlapType.TRIPLE_SLOTTED: 55.0,
            FlapType.FOWLER: 30.0,
            FlapType.DOUBLE_SLOTTED_FOWLER: 30.0,
        }

    def setup(self):

        aviary_options = self.options['aviary_options']

        self.add_subsystem(
            "BasicFlapsCalculations",
            BasicFlapsCalculations(),
            promotes_inputs=[
                "slat_defl",
                "flap_defl",
            ]
            + ["aircraft:*"],
            promotes_outputs=["*"],
        )

        self.add_subsystem(
            "CLmaxCalculation",
            CLmaxCalculation(),
            promotes_inputs=[
                Dynamic.Mission.SPEED_OF_SOUND,
                Dynamic.Mission.STATIC_PRESSURE,
                "kinematic_viscosity",
                "VLAM1",
                "VLAM2",
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM8",
                "VLAM9",
                "VLAM10",
                "VLAM11",
                "VLAM12",
                "VLAM13",
                "VLAM14",
                "fus_lift",
                Dynamic.Mission.TEMPERATURE,
            ]
            + ["aircraft:*"],
            promotes_outputs=["CL_max", Dynamic.Mission.MACH, "reynolds"],
        )

        self.add_subsystem(
            "LookupTables",
            MetaModelGroup(aviary_options=aviary_options),
            promotes_inputs=[
                "flap_defl_ratio",
                "flap_defl",
                "slat_defl_ratio",
                "reynolds",
                Dynamic.Mission.MACH,
                "body_to_span_ratio",
                "chord_to_body_ratio",
            ]
            + ["aircraft:*"],
            promotes_outputs=[
                "VDEL1",
                "VDEL2",
                "VDEL3",
                "VLAM1",
                "VLAM2",
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM10",
                "VLAM11",
                "VLAM13",
                "VLAM14",
                "fus_lift",
            ],
        )

        self.add_subsystem(
            "LiftAndDragIncrements",
            LiftAndDragIncrements(),
            promotes_inputs=[
                Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM,
                Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM,
                "VDEL1",
                "VDEL2",
                "VDEL3",
                "VDEL4",
                "VDEL5",
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM8",
                "VLAM13",
                "VLAM14",
            ],
            promotes_outputs=["delta_CD", "delta_CL"],
        )

        self.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
        self.linear_solver = om.DirectSolver()

        self.nonlinear_solver.options["iprint"] = 0
        self.nonlinear_solver.options["maxiter"] = 25
        self.nonlinear_solver.options["atol"] = 1e-8
        self.nonlinear_solver.options["rtol"] = 1e-8

        # set default trailing edge deflection angle per GASP
        self.set_input_defaults(
            Aircraft.Wing.OPTIMUM_FLAP_DEFLECTION,
            self.optimum_flap_defls[self.options["aviary_options"].get_val(
                Aircraft.Wing.FLAP_TYPE, units='unitless')],
            units="deg",
        )



import numpy as np
import openmdao.api as om

from aviary.subsystems.geometry.flops_based.utils import Names
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class CharacteristicLengths(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        self.add_input(Names.CROOT, 0.0, units='unitless')

        add_aviary_input(self, Aircraft.Canard.AREA, 0.0)
        add_aviary_input(self, Aircraft.Canard.ASPECT_RATIO, 0.0)
        # add_aviary_input(self, Aircraft.Canard.LAMINAR_FLOW_LOWER, 0.0)
        # add_aviary_input(self, Aircraft.Canard.LAMINAR_FLOW_UPPER, 0.0)
        add_aviary_input(self, Aircraft.Canard.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, 0.0)
        # add_aviary_input(self, Aircraft.Fuselage.LAMINAR_FLOW_LOWER, 0.0)
        # add_aviary_input(self, Aircraft.Fuselage.LAMINAR_FLOW_UPPER, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, 0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.AREA, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.ASPECT_RATIO, 0.0)
        # add_aviary_input(self, Aircraft.HorizontalTail.LAMINAR_FLOW_LOWER, 0.0)
        # add_aviary_input(self, Aircraft.HorizontalTail.LAMINAR_FLOW_UPPER, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(self, Aircraft.Nacelle.AVG_DIAMETER, 0.0)
        add_aviary_input(self, Aircraft.Nacelle.AVG_LENGTH, 0.0)
        # add_aviary_input(self, Aircraft.Nacelle.LAMINAR_FLOW_LOWER, 0.0)
        # add_aviary_input(self, Aircraft.Nacelle.LAMINAR_FLOW_UPPER, 0.0)

        add_aviary_input(self, Aircraft.VerticalTail.AREA, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.ASPECT_RATIO, 0.0)
        # add_aviary_input(self, Aircraft.VerticalTail.LAMINAR_FLOW_LOWER, 0.0)
        # add_aviary_input(self, Aircraft.VerticalTail.LAMINAR_FLOW_UPPER, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.THICKNESS_TO_CHORD, 0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.GLOVE_AND_BAT, 0.0)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)
        # add_aviary_input(self, Aircraft.Wing.LAMINAR_FLOW_LOWER, 0.0)
        # add_aviary_input(self, Aircraft.Wing.LAMINAR_FLOW_UPPER, 0.0)

        add_aviary_output(self, Aircraft.Canard.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_output(self, Aircraft.Canard.FINENESS, 0.0)

        add_aviary_output(self, Aircraft.Fuselage.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_output(self, Aircraft.Fuselage.FINENESS, 0.0)

        add_aviary_output(
            self, Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH, 0.0
        )

        add_aviary_output(self, Aircraft.HorizontalTail.FINENESS, 0.0)

        add_aviary_output(self, Aircraft.Nacelle.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_output(self, Aircraft.Nacelle.FINENESS, 0.0)

        add_aviary_output(
            self, Aircraft.VerticalTail.CHARACTERISTIC_LENGTH, 0.0
        )

        add_aviary_output(self, Aircraft.VerticalTail.FINENESS, 0.0)

        add_aviary_output(self, Aircraft.Wing.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_output(self, Aircraft.Wing.FINENESS, 0.0)

    def setup_partials(self):
        self._setup_partials_wing()
        self._setup_partials_horizontal_tail()
        self._setup_partials_vertical_tail()
        self._setup_partials_fuselage()
        self._setup_partials_nacelles()
        self._setup_partials_canard()

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        self._compute_wing(inputs, outputs, discrete_inputs, discrete_outputs)

        self._compute_horizontal_tail(
            inputs, outputs, discrete_inputs, discrete_outputs
        )

        self._compute_vertical_tail(
            inputs, outputs, discrete_inputs, discrete_outputs
        )

        self._compute_fuselage(
            inputs, outputs, discrete_inputs, discrete_outputs
        )

        self._compute_nacelles(
            inputs, outputs, discrete_inputs, discrete_outputs
        )

        # self._compute_additional_fuselages(
        #     inputs, outputs, discrete_inputs, discrete_outputs
        # )

        # self._compute_additional_vertical_tails(
        #     inputs, outputs, discrete_inputs, discrete_outputs
        # )

        self._compute_canard(
            inputs, outputs, discrete_inputs, discrete_outputs
        )

    def compute_partials(self, inputs, J, discrete_inputs=None):
        self._compute_partials_wing(inputs, J, discrete_inputs)
        self._compute_partials_horizontal_tail(inputs, J, discrete_inputs)
        self._compute_partials_vertical_tail(inputs, J, discrete_inputs)
        self._compute_partials_fuselage(inputs, J, discrete_inputs)
        self._compute_partials_nacelles(inputs, J, discrete_inputs)
        self._compute_partials_canard(inputs, J, discrete_inputs)

    def _setup_partials_wing(self):
        wrt = [
            Aircraft.Wing.AREA,
            Aircraft.Wing.ASPECT_RATIO,
            Aircraft.Wing.GLOVE_AND_BAT,
        ]

        aviary_options: AviaryValues = self.options['aviary_options']

        if aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION):
            wrt = [
                Names.CROOT,
                Aircraft.Wing.TAPER_RATIO,
            ]

        self.declare_partials(Aircraft.Wing.CHARACTERISTIC_LENGTH, wrt)

        self.declare_partials(
            Aircraft.Wing.FINENESS, Aircraft.Wing.THICKNESS_TO_CHORD, val=1.0
        )

    def _setup_partials_horizontal_tail(self):
        self.declare_partials(
            Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
            [
                Aircraft.HorizontalTail.AREA,
                Aircraft.HorizontalTail.ASPECT_RATIO,
            ]
        )

        self.declare_partials(
            Aircraft.HorizontalTail.FINENESS,
            Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=1.0
        )

    def _setup_partials_vertical_tail(self):
        self.declare_partials(
            Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
            [
                Aircraft.VerticalTail.AREA,
                Aircraft.VerticalTail.ASPECT_RATIO,
            ]
        )

        self.declare_partials(
            Aircraft.VerticalTail.FINENESS,
            Aircraft.VerticalTail.THICKNESS_TO_CHORD, val=1.0
        )

    def _setup_partials_fuselage(self):
        self.declare_partials(
            Aircraft.Fuselage.CHARACTERISTIC_LENGTH,
            Aircraft.Fuselage.LENGTH, val=1.0
        )

        self.declare_partials(
            Aircraft.Fuselage.FINENESS,
            [
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.Fuselage.LENGTH,
            ]
        )

    def _setup_partials_nacelles(self):
        self.declare_partials(
            Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
            Aircraft.Nacelle.AVG_LENGTH
        )

        self.declare_partials(
            Aircraft.Nacelle.FINENESS,
            [
                Aircraft.Nacelle.AVG_DIAMETER,
                Aircraft.Nacelle.AVG_LENGTH,
            ]
        )

    def _setup_partials_canard(self):
        self.declare_partials(
            Aircraft.Canard.CHARACTERISTIC_LENGTH,
            [
                Aircraft.Canard.AREA,
                Aircraft.Canard.ASPECT_RATIO,
            ]
        )

        self.declare_partials(
            Aircraft.Canard.FINENESS,
            Aircraft.Canard.THICKNESS_TO_CHORD,
        )

    def _compute_wing(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        area = inputs[Aircraft.Wing.AREA]
        glove_and_bat = inputs[Aircraft.Wing.GLOVE_AND_BAT]
        aspect_ratio = inputs[Aircraft.Wing.ASPECT_RATIO]

        length = ((area - glove_and_bat) / aspect_ratio)**0.5

        aviary_options: AviaryValues = self.options['aviary_options']

        if aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION):
            taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
            CROOT = inputs[Names.CROOT]

            length = (
                2.0 * CROOT * (1.0 + taper_ratio + taper_ratio**2.0)
                / (3.0 + 3.0 * taper_ratio)
            )

        outputs[Aircraft.Wing.CHARACTERISTIC_LENGTH] = length

        thickness_to_chord = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]

        outputs[Aircraft.Wing.FINENESS] = thickness_to_chord

    def _compute_horizontal_tail(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aspect_ratio = inputs[Aircraft.HorizontalTail.ASPECT_RATIO]

        length = 0.0

        if 0.0 < aspect_ratio:
            area = inputs[Aircraft.HorizontalTail.AREA]

            length = (area / aspect_ratio)**0.5

        outputs[Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH] = length

        thickness_to_chord = inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]

        outputs[Aircraft.HorizontalTail.FINENESS] = thickness_to_chord

    def _compute_vertical_tail(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aspect_ratio = inputs[Aircraft.VerticalTail.ASPECT_RATIO]

        length = 0.0

        if 0.0 < aspect_ratio:
            area = inputs[Aircraft.VerticalTail.AREA]

            length = (area / aspect_ratio)**0.5

        outputs[Aircraft.VerticalTail.CHARACTERISTIC_LENGTH] = length

        thickness_to_chord = inputs[Aircraft.VerticalTail.THICKNESS_TO_CHORD]

        outputs[Aircraft.VerticalTail.FINENESS] = thickness_to_chord

    def _compute_fuselage(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        length = inputs[Aircraft.Fuselage.LENGTH]

        outputs[Aircraft.Fuselage.CHARACTERISTIC_LENGTH] = length

        avg_diam = inputs[Aircraft.Fuselage.AVG_DIAMETER]

        fineness = length / avg_diam

        outputs[Aircraft.Fuselage.FINENESS] = fineness

    def _compute_nacelles(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        # TODO do all engines support nacelles? If not, is this deliberate, or
        # just an artifact of the implementation?
        num_eng = self.options['aviary_options'].get_val(Aircraft.Engine.NUM_ENGINES)

        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]

        char_len = np.zeros(len(num_eng), dtype=avg_diam.dtype)
        fineness = np.zeros(len(num_eng), dtype=avg_diam.dtype)

        num_idx = np.where(num_eng >= 1)
        char_len[num_idx] = avg_length[num_idx]
        fineness[num_idx] = 1.0

        calc_idx = np.intersect1d(np.where(avg_diam[num_idx] > 0), num_idx)

        if 0.0 < avg_diam[calc_idx]:
            fineness[calc_idx] = avg_length[calc_idx] / avg_diam[calc_idx]

        outputs[Aircraft.Nacelle.CHARACTERISTIC_LENGTH] = char_len
        outputs[Aircraft.Nacelle.FINENESS] = fineness

    def _compute_additional_fuselages(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        num_fuselages = inputs[Aircraft.Fuselage.NUM_FUSELAGES]

        if num_fuselages < 2:
            return

        num_extra = num_fuselages - 1

        idx = self._num_components
        self._num_components += num_extra

        lengths = outputs[Aircraft.Design.CHARACTERISTIC_LENGTHS]

        fineness = outputs[Aircraft.Design.FINENESS]

        laminar_flow_lower = outputs[Aircraft.Design.LAMINAR_FLOW_LOWER]
        laminar_flow_upper = outputs[Aircraft.Design.LAMINAR_FLOW_UPPER]

        for _ in range(num_extra):
            lengths[idx] = lengths[3]

            fineness[idx] = fineness[3]

            laminar_flow_lower[idx] = laminar_flow_lower[3]
            laminar_flow_upper[idx] = laminar_flow_upper[3]

            idx += 1

    def _compute_additional_vertical_tails(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_tails = aviary_options.get_val(Aircraft.VerticalTail.NUM_TAILS)

        if num_tails < 2:
            return

        num_extra = num_tails - 1

        idx = self._num_components
        self._num_components += num_extra

        lengths = outputs[Aircraft.Design.CHARACTERISTIC_LENGTHS]

        fineness = outputs[Aircraft.Design.FINENESS]

        laminar_flow_lower = outputs[Aircraft.Design.LAMINAR_FLOW_LOWER]
        laminar_flow_upper = outputs[Aircraft.Design.LAMINAR_FLOW_UPPER]

        for _ in range(num_extra):
            lengths[idx] = lengths[2]

            fineness[idx] = fineness[2]

            laminar_flow_lower[idx] = laminar_flow_lower[2]
            laminar_flow_upper[idx] = laminar_flow_upper[2]

            idx += 1

    def _compute_canard(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        area = inputs[Aircraft.Canard.AREA]

        if area <= 0.0:
            return

        thickness_to_chord = inputs[Aircraft.Canard.THICKNESS_TO_CHORD]
        aspect_ratio = inputs[Aircraft.Canard.ASPECT_RATIO]

        length = 0.0

        if 0.0 < aspect_ratio:
            length = (area / aspect_ratio)**0.5

        outputs[Aircraft.Canard.CHARACTERISTIC_LENGTH] = length

        outputs[Aircraft.Canard.FINENESS] = thickness_to_chord

    def _compute_partials_wing(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']

        if aviary_options.get_val(Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION):
            taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
            CROOT = inputs[Names.CROOT]

            a = 2.0 * (1.0 + taper_ratio + taper_ratio**2.0)
            f = a * CROOT
            df = 2.0 * CROOT * (1.0 + 2.0 * taper_ratio)
            g = (3.0 + 3.0 * taper_ratio)
            dg = 3.0

            J[
                Aircraft.Wing.CHARACTERISTIC_LENGTH,
                Names.CROOT
            ] = a / g

            J[
                Aircraft.Wing.CHARACTERISTIC_LENGTH,
                Aircraft.Wing.TAPER_RATIO
            ] = (df * g - f * dg) / g**2

        else:
            area = inputs[Aircraft.Wing.AREA]
            glove_and_bat = inputs[Aircraft.Wing.GLOVE_AND_BAT]
            aspect_ratio = inputs[Aircraft.Wing.ASPECT_RATIO]

            a = area - glove_and_bat
            f = 0.5 * (a / aspect_ratio)**-0.5
            df = f / aspect_ratio

            J[
                Aircraft.Wing.CHARACTERISTIC_LENGTH,
                Aircraft.Wing.AREA
            ] = df

            J[
                Aircraft.Wing.CHARACTERISTIC_LENGTH,
                Aircraft.Wing.GLOVE_AND_BAT
            ] = -df

            J[
                Aircraft.Wing.CHARACTERISTIC_LENGTH,
                Aircraft.Wing.ASPECT_RATIO
            ] = -f * a / aspect_ratio**2.0

    def _compute_partials_horizontal_tail(
        self, inputs, J, discrete_inputs=None
    ):
        aspect_ratio = inputs[Aircraft.HorizontalTail.ASPECT_RATIO]

        da = dr = 0.0

        if 0.0 < aspect_ratio:
            area = inputs[Aircraft.HorizontalTail.AREA]

            f = 0.5 * (area / aspect_ratio)**-0.5
            da = f / aspect_ratio
            dr = -f * area / aspect_ratio**2.0

        J[
            Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
            Aircraft.HorizontalTail.AREA
        ] = da

        J[
            Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
            Aircraft.HorizontalTail.ASPECT_RATIO
        ] = dr

    def _compute_partials_vertical_tail(self, inputs, J, discrete_inputs=None):
        aspect_ratio = inputs[Aircraft.VerticalTail.ASPECT_RATIO]

        da = dr = 0.0

        if 0.0 < aspect_ratio:
            area = inputs[Aircraft.VerticalTail.AREA]

            f = 0.5 * (area / aspect_ratio)**-0.5
            da = f / aspect_ratio
            dr = -f * area / aspect_ratio**2.0

        J[
            Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
            Aircraft.VerticalTail.AREA
        ] = da

        J[
            Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
            Aircraft.VerticalTail.ASPECT_RATIO
        ] = dr

    def _compute_partials_fuselage(self, inputs, J, discrete_inputs=None):
        length = inputs[Aircraft.Fuselage.LENGTH]
        avg_diam = inputs[Aircraft.Fuselage.AVG_DIAMETER]

        J[
            Aircraft.Fuselage.FINENESS,
            Aircraft.Fuselage.LENGTH
        ] = 1.0 / avg_diam

        J[
            Aircraft.Fuselage.FINENESS,
            Aircraft.Fuselage.AVG_DIAMETER
        ] = -length / avg_diam**2.0

    def _compute_partials_nacelles(self, inputs, J, discrete_inputs=None):
        num_eng = self.options['aviary_options'].get_val(Aircraft.Engine.NUM_ENGINES)

        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]

        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]

        deriv_char_len = np.zeros(len(num_eng), dtype=avg_diam.dtype)
        deriv_fine_len = np.zeros(len(num_eng), dtype=avg_diam.dtype)
        deriv_fine_diam = np.zeros(len(num_eng), dtype=avg_diam.dtype)

        calc_idx = np.where(num_eng >= 1)
        deriv_char_len[calc_idx] = 1.0
        deriv_fine_len[calc_idx] = 1.0 / avg_diam[calc_idx]
        deriv_fine_diam[calc_idx] = -avg_length[calc_idx] / avg_diam[calc_idx]**2.0

        J[
            Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
            Aircraft.Nacelle.AVG_LENGTH
        ] = deriv_char_len

        J[
            Aircraft.Nacelle.FINENESS,
            Aircraft.Nacelle.AVG_LENGTH
        ] = deriv_fine_len

        J[
            Aircraft.Nacelle.FINENESS,
            Aircraft.Nacelle.AVG_DIAMETER
        ] = deriv_fine_diam

    def _compute_partials_canard(self, inputs, J, discrete_inputs=None):
        area = inputs[Aircraft.Canard.AREA]

        if area <= 0.0:
            J[
                Aircraft.Canard.CHARACTERISTIC_LENGTH,
                Aircraft.Canard.AREA
            ] = J[
                Aircraft.Canard.CHARACTERISTIC_LENGTH,
                Aircraft.Canard.ASPECT_RATIO
            ] = J[
                Aircraft.Canard.FINENESS,
                Aircraft.Canard.THICKNESS_TO_CHORD
            ] = 0.0

            return

        aspect_ratio = inputs[Aircraft.Canard.ASPECT_RATIO]

        da = dr = 0.0

        if 0.0 < aspect_ratio:
            area = inputs[Aircraft.Canard.AREA]

            f = 0.5 * (area / aspect_ratio)**-0.5
            da = f / aspect_ratio
            dr = -f * area / aspect_ratio**2.0

        J[
            Aircraft.Canard.CHARACTERISTIC_LENGTH,
            Aircraft.Canard.AREA
        ] = da

        J[
            Aircraft.Canard.CHARACTERISTIC_LENGTH,
            Aircraft.Canard.ASPECT_RATIO
        ] = dr

        J[
            Aircraft.Canard.FINENESS,
            Aircraft.Canard.THICKNESS_TO_CHORD
        ] = 1.0



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.mux_component import MuxComponent
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft


class TestMuxComponent(unittest.TestCase):

    def test_mux(self):
        prob = om.Problem()
        model = prob.model

        aviary_options = AviaryValues()
        aviary_options.set_val(Aircraft.VerticalTail.NUM_TAILS, 1)
        aviary_options.set_val(Aircraft.Fuselage.NUM_FUSELAGES, 1)
        aviary_options.set_val(Aircraft.Engine.NUM_ENGINES, np.array([2]))

        model.add_subsystem(
            'mux', MuxComponent(aviary_options=aviary_options),
            promotes_inputs=['*'])

        prob.setup(force_alloc_complex=True)

        prob.set_val(Aircraft.Wing.WETTED_AREA, 2.0)
        prob.set_val(Aircraft.Wing.FINENESS, 3.0)
        prob.set_val(Aircraft.Wing.CHARACTERISTIC_LENGTH, 4.0)
        prob.set_val(Aircraft.Wing.LAMINAR_FLOW_UPPER, 0.01)
        prob.set_val(Aircraft.Wing.LAMINAR_FLOW_LOWER, 0.02)

        prob.set_val(Aircraft.HorizontalTail.WETTED_AREA, 12.0)
        prob.set_val(Aircraft.HorizontalTail.FINENESS, 13.0)
        prob.set_val(Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH, 14.0)
        prob.set_val(Aircraft.HorizontalTail.LAMINAR_FLOW_UPPER, 0.03)
        prob.set_val(Aircraft.HorizontalTail.LAMINAR_FLOW_LOWER, 0.04)

        prob.set_val(Aircraft.VerticalTail.WETTED_AREA, 22.0)
        prob.set_val(Aircraft.VerticalTail.FINENESS, 23.0)
        prob.set_val(Aircraft.VerticalTail.CHARACTERISTIC_LENGTH, 24.0)
        prob.set_val(Aircraft.VerticalTail.LAMINAR_FLOW_UPPER, 0.05)
        prob.set_val(Aircraft.VerticalTail.LAMINAR_FLOW_LOWER, 0.06)

        prob.set_val(Aircraft.Fuselage.WETTED_AREA, 32.0)
        prob.set_val(Aircraft.Fuselage.FINENESS, 33.0)
        prob.set_val(Aircraft.Fuselage.CHARACTERISTIC_LENGTH, 34.0)
        prob.set_val(Aircraft.Fuselage.LAMINAR_FLOW_UPPER, 0.07)
        prob.set_val(Aircraft.Fuselage.LAMINAR_FLOW_LOWER, 0.08)

        prob.set_val(Aircraft.Nacelle.WETTED_AREA, np.array([42.0]))
        prob.set_val(Aircraft.Nacelle.FINENESS, np.array([43.0]))
        prob.set_val(Aircraft.Nacelle.CHARACTERISTIC_LENGTH, np.array([44.0]))
        prob.set_val(Aircraft.Nacelle.LAMINAR_FLOW_UPPER, np.array([0.09]))
        prob.set_val(Aircraft.Nacelle.LAMINAR_FLOW_LOWER, np.array([0.10]))

        prob.run_model()
        prob.check_partials(compact_print=True, method="cs")

        derivs = prob.check_partials(out_stream=None, method="cs")

        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)

        # Truth values for each variable
        wetted_areas_truth = np.array([2., 12., 22., 32., 42., 42.])
        fineness_ratios_truth = np.array([3., 13., 23., 33., 43., 43.])
        characteristic_lengths_truth = np.array([4., 14., 24., 34., 44., 44.])
        laminar_fractions_upper_truth = np.array([0.01, 0.03, 0.05, 0.07, 0.09, 0.09])
        laminar_fractions_lower_truth = np.array([0.02, 0.04, 0.06, 0.08, 0.1, 0.1])

        wetted_areas_output = prob['mux.wetted_areas']
        fineness_ratios_output = prob['mux.fineness_ratios']
        characteristic_lengths_output = prob['mux.characteristic_lengths']
        laminar_fractions_upper_output = prob['mux.laminar_fractions_upper']
        laminar_fractions_lower_output = prob['mux.laminar_fractions_lower']

        # Assert that the outputs are near the expected values
        assert_near_equal(wetted_areas_output, wetted_areas_truth, 1e-7)
        assert_near_equal(fineness_ratios_output, fineness_ratios_truth, 1e-7)
        assert_near_equal(characteristic_lengths_output,
                          characteristic_lengths_truth, 1e-7)
        assert_near_equal(laminar_fractions_upper_output,
                          laminar_fractions_upper_truth, 1e-7)
        assert_near_equal(laminar_fractions_lower_output,
                          laminar_fractions_lower_truth, 1e-7)


if __name__ == "__main__":
    unittest.main()




import warnings
from enum import Enum

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from parameterized import param

from aviary.utils.aviary_values import AviaryValues
from aviary.utils.options import list_options as list_options_func
from aviary.utils.preprocessors import preprocess_options
from aviary.validation_cases.validation_data.flops_data.FLOPS_Test_Data import \
    FLOPS_Test_Data
from aviary.variable_info.variables import Aircraft

Version = Enum('Version', ['ALL', 'TRANSPORT', 'ALTERNATE', 'BWB'])


def do_validation_test(prob: om.Problem,
                       case_name: str,
                       input_validation_data: AviaryValues,
                       output_validation_data: AviaryValues,
                       input_keys: list,
                       output_keys: list,
                       aviary_option_keys: list = None,
                       tol: float = 1.0e-4,
                       atol: float = 1.0e-12,
                       rtol: float = 1.0e-12,
                       method: str = "cs",
                       step: float = None,
                       check_values: bool = True,
                       check_partials: bool = True,
                       excludes: list = None,
                       list_inputs: bool = False,
                       list_outputs: bool = False,
                       list_options: bool = False):
    """
    Runs a validation test with user-supplied validation data.

    Parameters
    ----------
    prob : om.Problem
        An instantiated and set-up Problem.
    case_name : str
        Name of the case being run.
    input_validation_data : AviaryValues
        Input variables and their values to use when running the problem.
        The object must contain at a minimum all the input variables listed
        in input_keys.
    output_validation_data : AviaryValues
        Output variables and their values to which output from the problem
        will be compared. The object must contain at a minimum all the
        output variables listed in output_keys. Note that input_validation_data
        and output_validation_data may be the same object.
    input_keys : str, or iter of str
        List of input variables whose values will be transferred from
        the input validation data in order to run the problem.
    output_keys : str, or iter of str
        List of output variables whose values will be looked up in
        the output validation data and compared against the outputs from the
        problem.
    aviary_option_keys: str, or iter of str
        List of aviary_options keys whose values will be looked up and
        listed in the options printout. If None, all items in aviary_options
        will be listed.
    tol : float
        Relative tolerance for comparing problem outputs against
        validation data. The default is 1.0e-4.
    atol : float
        Absolute tolerance for checking partial derivative calculations. The
        default is 1.0e-12.
    rtol : float
        Relative tolerance for checking partial derivative calculations. The
        default is 1.0e-12.
    method : str
        Type of differencing to use. The default is "cs".
    step : float
        Step size for approximation. Default is None, which means 1e-6 for 'fd' and 1e-40 for
        'cs'.
    check_values : bool
        If true, check output values against validation data. The default is
        true.
    check_partials : bool
        If true, check partial derivative calculations. The default is true.
    excludes : None or list_like
        List of glob patterns for pathnames to exclude from the partial derivative check.
        Default is None, which excludes nothing.
    list_options: bool
        If True, values of options for all components in the model will be listed
        on standard output. Default is False.
    list_inputs: bool
        If True, prob.model.list_inputs() will be called after the model is run.
        Default is False.
    list_outputs: bool
        If True, prob.model.list_outputs() will be called after the model is run.
        Default is False.
    """

    input_key_list = _assure_is_list(input_keys)
    output_key_list = _assure_is_list(output_keys)
    aviary_option_key_list = _assure_is_list(aviary_option_keys)

    for key in input_key_list:
        if key in input_validation_data:
            data = input_validation_data
        else:
            data = output_validation_data
        desired, units = data.get_item(key)
        prob.set_val(key, desired, units)

    prob.run_model()
    print(case_name)

    if list_options:
        list_options_func(prob.model, aviary_keys=aviary_option_key_list)

    if list_inputs:
        prob.model.list_inputs()

    if list_outputs:
        prob.model.list_outputs()

    if check_values:
        for key in output_key_list:
            desired, units = output_validation_data.get_item(key)
            actual = prob.get_val(key, units)
            try:
                assert_near_equal(actual, desired, tol)
            except ValueError as err:
                raise ValueError(f'ValueError for key = {key}') from err

    if check_partials:
        partial_data = prob.check_partials(
            compact_print=True, method=method, step=step, excludes=excludes)
        assert_check_partials(partial_data, atol=atol, rtol=rtol)


def flops_validation_test(prob: om.Problem,
                          case_name: str,
                          input_keys: list,
                          output_keys: list,
                          aviary_option_keys: list = None,
                          version: Version = Version.ALL,
                          tol: float = 1.0e-4,
                          atol: float = 1.0e-12,
                          rtol: float = 1.0e-12,
                          method: str = "cs",
                          step: float = None,
                          check_values: bool = True,
                          check_partials: bool = True,
                          excludes: list = None,
                          list_inputs: bool = False,
                          list_outputs: bool = False,
                          list_options: bool = False,
                          flops_inputs=None,
                          flops_outputs=None):
    """
    Runs a validation test using FLOPS validation data.

    Parameters
    ----------
    prob : om.Problem
        An instantiated and set-up Problem.
    case_name : str
        Name of the case being run. Validation data will be looked up from
        the corresponding case in the FLOPS validation data collection.
    input_keys : str, or iter of str
        List of input variables whose values will be transferred from
        the input validation data in order to run the problem.
    output_keys : str, or iter of str
        List of output variables whose values will be looked up in
        the output validation data and compared against the outputs from the
        problem.
    aviary_option_keys: str, or iter of str
        List of aviary_options keys whose values will be looked up and
        listed in the options printout. If None, all items in aviary_options
        will be listed.
    version: Version
        If this is a FLOPS-based mass analysis test, version specifies which
        version of the mass equations being tested. Currently, there is no
        BWB validation data so output values will not be checked.
        Default is ALL.
    tol : float
        Relative tolerance for comparing problem outputs against
        validation data. The default is 1.0e-4.
    atol : float
        Absolute tolerance for checking partial derivative calculations. The
        default is 1.0e-12.
    rtol : float
        Relative tolerance for checking partial derivative calculations. The
        default is 1.0e-12.
    method : str
        Type of differencing to use. The default is "cs".
    check_values : bool
        If true, check output values against validation data. The default is
        true. This option may be overridden in cases where validation data are
        not available; currently this applies to BWB aircraft.
    check_partials : bool
        If true, check partial derivative calculations. The default is true.
    excludes : None or list_like
        List of glob patterns for pathnames to exclude from the partial derivative check.
        Default is None, which excludes nothing.
    list_options: bool
        If True, values of options for all components in the model will be listed
        on standard output. Default is False.
    list_inputs: bool
        If True, prob.model.list_inputs() will be called after the model is run.
        Default is False.
    list_outputs: bool
        If True, prob.model.list_outputs() will be called after the model is run.
        Default is False.
    flops_inputs: None or AviaryValues
        Allows a custom set of inputs to be tested. Default is None, which reads
        data from FLOPS_Test_Data with key case_name.
    flops_outputs: None or AviaryValues
        Allows a custom set of outputs to be tested. Default is None, which reads
        data from FLOPS_Test_Data with key case_name.
    """

    if not isinstance(version, Version):
        raise TypeError('parameter "version" must be of enumeration type "Version"')

    if flops_inputs is None and flops_outputs is None:
        flops_data = FLOPS_Test_Data[case_name]
        flops_inputs = flops_data['inputs']
        flops_outputs = flops_data['outputs']

    if version is Version.TRANSPORT and flops_inputs.get_val(Aircraft.Design.USE_ALT_MASS) or \
            version is Version.ALTERNATE and not flops_inputs.get_val(Aircraft.Design.USE_ALT_MASS):
        return

    # TODO: Currently no BWB validation data.
    # For BWBs, skip the validation test, but do check the partials.
    check_values = (
        check_values and version is not Version.BWB
    )
    if not check_values:
        warnings.warn('Not checking values because validation data not available.')

    do_validation_test(prob=prob,
                       case_name=case_name,
                       input_validation_data=flops_inputs,
                       output_validation_data=flops_outputs,
                       input_keys=input_keys,
                       output_keys=output_keys,
                       aviary_option_keys=aviary_option_keys,
                       tol=tol,
                       atol=atol,
                       rtol=rtol,
                       method=method,
                       step=step,
                       check_values=check_values,
                       check_partials=check_partials,
                       excludes=excludes,
                       list_options=list_options,
                       list_inputs=list_inputs,
                       list_outputs=list_outputs)


def get_flops_data(case_name: str, keys: str = None) -> AviaryValues:
    """
    Returns an AviaryValues object containing input and output data for
    the named FLOPS validation case.

    Parameters
    ----------
    case_name : str
        Name of the case being run. Validation data will be looked up from
        the corresponding case in the FLOPS validation data collection.
    keys : str, or iter of str
        List of variables whose values will be transferred from the validation data.
        The default is all variables.
    """
    flops_data_copy: AviaryValues = get_flops_inputs(case_name)
    flops_data_copy.update(get_flops_outputs(case_name))
    if keys is None:
        return flops_data_copy
    keys_list = _assure_is_list(keys)

    return AviaryValues({key: flops_data_copy.get_item(key) for key in keys_list})


def get_flops_inputs(case_name: str, keys: str = None, preprocess: bool = False) -> AviaryValues:
    """
    Returns an AviaryValues object containing input data for the named FLOPS validation case.

    Parameters
    ----------
    case_name : str
        Name of the case being run. Input data will be looked up from
        the corresponding case in the FLOPS validation data collection.
    keys : str, or iter of str
        List of variables whose values will be transferred from the input data.
        The default is all variables.
    preprocess: bool
        If true, the input data will be passed through preprocess_options() to
        fill in any missing options before being returned. The default is False.
    """
    flops_data: dict = FLOPS_Test_Data[case_name]
    flops_inputs_copy: AviaryValues = flops_data['inputs'].deepcopy()
    if preprocess:
        preprocess_options(flops_inputs_copy)
    if keys is None:
        return flops_inputs_copy
    keys_list = _assure_is_list(keys)

    return AviaryValues({key: flops_inputs_copy.get_item(key) for key in keys_list})


def get_flops_outputs(case_name: str, keys: str = None) -> AviaryValues:
    """
    Returns an AviaryValues object containing output data for the named FLOPS validation case.

    Parameters
    ----------
    case_name : str
        Name of the case being run. Output data will be looked up from
        the corresponding case in the FLOPS validation data collection.
    keys : str, or iter of str
        List of variables whose values will be transferred from the output data.
        The default is all variables.
    """
    flops_data: dict = FLOPS_Test_Data[case_name]
    flops_outputs_copy: AviaryValues = flops_data['outputs'].deepcopy()
    if keys is None:
        return flops_outputs_copy
    keys_list = _assure_is_list(keys)

    return AviaryValues({key: flops_outputs_copy.get_item(key) for key in keys_list})


def get_flops_case_names(omit: list = None, only: list = None) -> list:
    """
    Returns a list of case names in the FLOPS validation database.

    Parameters
    ----------
    omit : str or list
        The case name or list of case names to omit from the returned list.
    only : str or list
        The case name or list of case names to be considered for inclusion
        in the returned list. Note that parameters only and omit cannot be
        used together.
    """
    if omit is not None and only is not None:
        raise ValueError('cannot use both only and omit keywords')

    omit_list = _assure_is_list(omit)
    only_list = _assure_is_list(only, backup=FLOPS_Test_Data)

    return [key for key in FLOPS_Test_Data if key in only_list and key not in omit_list]


def print_case(testcase_func, param_num, param: param):
    """
    Returns a formatted case name for unit testing with decorator @parameterized.expand().
    It is intended to be used when expand() is called with a list of strings
    representing test case names.

    Parameters
    ----------
    testcase_func : Any
        This parameter is ignored.
    param_num : Any
        This parameter is ignored.
    param : param
        The param object containing the case name to be formatted.
    """
    return 'test_case_' + param.args[0]


def _assure_is_list(keys, backup=None):

    if isinstance(keys, str):
        return [keys]
    elif not keys:
        if backup is None:
            backup = []
        return backup
    return keys



import openmdao.api as om

from aviary.subsystems.mass.gasp_based.design_load import DesignLoadGroup
from aviary.subsystems.mass.gasp_based.equipment_and_useful_load import \
    EquipAndUsefulLoadMass
from aviary.subsystems.mass.gasp_based.fixed import FixedMassGroup
from aviary.subsystems.mass.gasp_based.fuel import FuelMassGroup
from aviary.subsystems.mass.gasp_based.wing import WingMassGroup
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft


class MassPremission(om.Group):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        # output values from design_load that are connected to fixed_mass via promotion
        fixed_mass_design_load_values = [
            "max_mach", "min_dive_vel"]

        # output values from fixed_mass that are connected to wing_mass via promotion
        wing_mass_fixed_mass_values = [
            "c_strut_braced",
            "c_gear_loc",
            "half_sweep",
        ]

        # output values from design_load that are connected to fuel_mass via promotion
        fuel_mass_design_load_values = ["min_dive_vel"]

        # output values from fixed_mass that are connected to fuel_mass via promotion
        fuel_mass_fixed_mass_values = [
            "payload_mass_des",
            "payload_mass_max",
            "wing_mounted_mass",
            "eng_comb_mass",
        ]

        # combine all necessary inputs and outputs for each group

        design_load_outputs = fixed_mass_design_load_values

        fixed_mass_inputs = fixed_mass_design_load_values + ["density"]
        fixed_mass_outputs = wing_mass_fixed_mass_values + fuel_mass_fixed_mass_values

        wing_mass_inputs = wing_mass_fixed_mass_values

        fuel_mass_inputs = (
            fuel_mass_design_load_values
            + fuel_mass_fixed_mass_values
        )

        # create the instances of the groups

        self.add_subsystem(
            "design_load",
            DesignLoadGroup(
                aviary_options=aviary_options,
            ),
            promotes_inputs=["aircraft:*", "mission:*"],
            promotes_outputs=["*"],
        )

        self.add_subsystem(
            "fixed_mass",
            FixedMassGroup(
                aviary_options=aviary_options,
            ),
            promotes_inputs=fixed_mass_inputs + ["aircraft:*", "mission:*"],
            promotes_outputs=fixed_mass_outputs + ["aircraft:*"],
        )

        self.add_subsystem(
            "equip_and_useful_mass",
            EquipAndUsefulLoadMass(
                aviary_options=aviary_options,
            ),
            promotes_inputs=["aircraft:*", "mission:*"],
            promotes_outputs=["aircraft:*"],
        )

        self.add_subsystem(
            "wing_mass",
            WingMassGroup(
                aviary_options=aviary_options,
            ),
            promotes_inputs=wing_mass_inputs + ["aircraft:*", "mission:*"],
            promotes_outputs=["aircraft:*"],
        )

        self.add_subsystem(
            "fuel_mass",
            FuelMassGroup(aviary_options=aviary_options),
            promotes_inputs=fuel_mass_inputs + ["aircraft:*", "mission:*"],
            promotes_outputs=[
                "aircraft:*", "mission:*"
            ],
        )

        self.set_input_defaults(Aircraft.Fuselage.LENGTH, 200, 'ft')
        self.set_input_defaults(Aircraft.HorizontalTail.AREA, 20, 'ft**2')
        self.set_input_defaults(Aircraft.VerticalTail.AREA, 20, 'ft**2')
